"""

This file initializes the coolest module
and provides some basic information about the package.

"""

# Set the package release version
version_info = (0, 2, 3)
__version__ = '.'.join(str(c) for c in version_info)

# Set the package details
__author__ = 'Aymeric Galan, Austin Peel, Giorgos Vernardos & Herculens contributors'
__email__ = 'aymeric.galan@gmail.com'
__year__ = '2021-2025'
__credits__ = 'Herculens contributors, EPFL (see AUTHORS.md)'
__url__ = 'https://github.com/Herculens/herculens'
__description__ = 'Auto-differentiable strong lens modelling'
__license__ = 'MIT'
